package framework;

import Controller.HttpService;
import connection.DatagramConnection;
import connection.NetInformation;
import framework.handler.HandlerDictionary;
import framework.integration.SimulationAgent;
import framework.integration.SimulationAgentWrapper;
import framework.integration.simulation.Breakpoint;
import framework.integration.simulation.SimulationStateListener;
import framework.serverState.WatcherManager;
import framework.services.ServicesManager;
import framework.services.avaliableServices.NotificationService;
import framework.services.servicesFinder.ReflectionServiceFinder;
import model.handler.DiscoveryProtocolHandler;
import model.proccesor.RequestProcessor;
import protocol.ClientSideDiscoverProtocol;

import java.net.URI;
import java.util.Date;

import static exception.ExceptionHandler.runSafe;
import static framework.serverState.ServerState.state;

public abstract class SGIFramework implements SimulationAgent{
    private HttpService httpService;
    private WatcherManager watcherManager;
    private Date lastDate;

    public SGIFramework(String label, Date initialDate, Date finalDate) {
        runSafe(() -> start(label, new URI("")));
        state().setDates(initialDate, finalDate);
        lastDate = initialDate;
        watcherManager = new WatcherManager(this::valueGet);
    }

    private void start(String simulationName, URI servicesPath) throws Exception {
        NetInformation result = new DiscoveryProtocolHandler(new ClientSideDiscoverProtocol(simulationName), new DatagramConnection(35001)).start();
        httpService = new HttpService(new RequestProcessor(new HandlerDictionary(new SimulationAgentWrapper(this))), result.getPort());
        setupFramework(simulationName, servicesPath, result);
        httpService.start();
    }

    private void setupFramework(String simulationName, URI servicesPath, NetInformation result) {
        httpService.enableRequestQueue();
        state().setSimulationName(simulationName);
        state().setRemoteAddress(result.getRemoteIP());
        state().setSimulationStateListener(new SimulationStateListener());
        System.out.println("port assigned: " + result.getPort());
        ServicesManager.setUp(new ReflectionServiceFinder(), servicesPath.getPath());
    }

    public void refresh(Date date) {
        httpService.executeStoreRequests();
        if(isPaused(date)) {
            treatBreakpoints(date);
            watcherManager.sendWatcherValues(date);
            sendPercentageLoaded(date);
        }
    }

    private void treatBreakpoints(Date date) {
        if (state().breakpoints().haveBreakpointNotPassed(date)) {
            ServicesManager.get(NotificationService.class).broadcast("{\"type\":\"Breakpoint\"}");
            markBreakpointPassed(date);
            pause();
        }
    }

    private void markBreakpointPassed(Date date) {
        state().breakpoints().getBreakpointsNotPassed(date).stream()
                .forEach(Breakpoint::passed);
    }

    private boolean isPaused(Date date) {
        boolean isPaused = date.after(lastDate);
        lastDate = date;
        return isPaused;
    }

    private void sendPercentageLoaded(Date date) {
        Float percentage = calculatePercentage(state().getFinalDate(), state().getInitialDate(), date);
        ServicesManager.get(NotificationService.class).broadcast("{\"type\":\"progress\",\"info\":{\"percentage\":" + percentage + "}}");//TODO
    }

    private Float calculatePercentage(Date finalDate, Date initialDate, Date date) {
        if(date.after(initialDate)) {
            Float percentage = (float)(date.getTime() - initialDate.getTime())/(finalDate.getTime()-initialDate.getTime());
            return percentage * 100;
        }
        return 0F;
    }

    public void waitForRequest(boolean wait) {
        if(wait)
            httpService.enableRequestQueue();
        else
            httpService.disableRequestQueue();
    }

}
