package framework;

import framework.integration.SimulationAgent;
import framework.integration.SimulationAgentWrapper;
import framework.integration.simulation.Breakpoint;
import framework.integration.simulation.SimulationState;
import framework.integration.simulation.SimulationStateListener;
import framework.integration.simulation.Watcher;
import model.conection.HttpRequest;
import model.conection.Request;
import model.conection.Response;
import org.junit.Before;
import org.junit.Test;

import java.util.Date;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import static exception.ExceptionHandler.getSafe;
import static framework.serverState.ServerState.state;
import static org.junit.Assert.*;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.verify;

public class FrameworkIntegrationArchitecture {

    public static final String ROOTS = "[{\"name\":\"happy-building\",\"objectId\":\"happy-building\"},{\"name\":\"bored-building\",\"objectId\":\"bored-building\"},{\"name\":\"unhappy-building\",\"objectId\":\"unhappy-building\"}]";
    private static final String CHILDREN = "[{\"name\":\"room1\",\"objectId\":\"room1\"},{\"name\":\"room2\",\"objectId\":\"room2\"},{\"name\":\"room3\",\"objectId\":\"room3\"}]";
    private static final String SENSORS = "[{\"name\":\"sensor1\",\"objectId\":\"sensor1\"},{\"name\":\"sensor2\",\"objectId\":\"sensor2\"},{\"name\":\"sensor3\",\"objectId\":\"sensor3\"}]";
    private static final String IMPOSIBLE_STOP = "{\"error\":\"impossible reset simulation\"}";
    private static final String IMPOSIBLE_PAUSE = "{\"error\":\"impossible pause simulation\"}";
    private static final String IMPOSIBLE_PLAY = "{\"error\":\"impossible play simulation\"}";

    private SimulationAgent agent;
    private SimulationAgentWrapper wrapper;

    @Before
    public void setUp() throws Exception {
        agent = new MySimulationAgent();
        wrapper = new SimulationAgentWrapper(agent);
    }

    @Test
    public void shouldReturnTheRootsOnSimulationAgent() {
        Request request= getSafe(()->new HttpRequest("GET", "/RootObjects", ""));
        Response response = wrapper.execute(request);
        assertEquals(200, response.getStatusCode());
        assertEquals(ROOTS, response.getBody());
    }

    @Test
    public void shouldReturnTheChildrenOf() {
        Request request= getSafe(()->new HttpRequest("GET", "/ObjectComponents", "{\"objectId\":\"happy-building\"}"));
        Response response = wrapper.execute(request);
        assertEquals(200, response.getStatusCode());
        assertEquals(CHILDREN, response.getBody());
    }

    @Test
    public void shouldReturnTheAttributes() {
        Request request= getSafe(()->new HttpRequest("GET", "/MeasurableAttributes", "{\"objectId\":\"happy-building\"}"));
        Response response = wrapper.execute(request);
        assertEquals(200, response.getStatusCode());
        assertEquals(SENSORS, response.getBody());
    }

    @Test
    public void shouldResponseWithImpossibleStop() {
        Request request= getSafe(()->new HttpRequest("GET", "/Stop", "{\"objectId\":\"happy-building\"}"));
        Response response = wrapper.execute(request);
        assertEquals(500, response.getStatusCode());
        assertEquals(IMPOSIBLE_STOP, response.getBody());
    }

    @Test
    public void shouldResponseWithImpossiblePlay() {
        Request request= getSafe(()->new HttpRequest("GET", "/Play", "{\"objectId\":\"happy-building\"}"));
        Response response = wrapper.execute(request);
        assertEquals(500, response.getStatusCode());
        assertEquals(IMPOSIBLE_PLAY, response.getBody());
    }

    @Test
    public void shouldResponseWithImpossiblePause() {
        Request request= getSafe(()->new HttpRequest("GET", "/Pause", "{\"objectId\":\"happy-building\"}"));
        Response response = wrapper.execute(request);
        assertEquals(500, response.getStatusCode());
        assertEquals(IMPOSIBLE_PAUSE, response.getBody());
    }

    @Test
    public void shouldDetectThatBreakpointHaveBeenNotFired() {
        Date date = new Date(System.currentTimeMillis());
        state().breakpoints().add(Breakpoint.create(date));
        assertFalse(state().breakpoints().haveBreakpointNotPassed(date));
    }

    @Test
    public void shouldDetectThatBreakpointHaveBeenFired() {
        Date date = new Date(System.currentTimeMillis());
        Date dateAfter1Millisecond = new Date(System.currentTimeMillis()+1);
        state().breakpoints().add(Breakpoint.create(date));
        assertTrue(state().breakpoints().haveBreakpointNotPassed(dateAfter1Millisecond));
    }

    @Test
    public void shouldDetectThatBreakpointHaveBeenNotFiredTimeBefore() {
        Date date = new Date(System.currentTimeMillis());
        Date dateAfter1Millisecond = new Date(System.currentTimeMillis()-1);
        state().breakpoints().add(Breakpoint.create(date));
        assertFalse(state().breakpoints().haveBreakpointNotPassed(dateAfter1Millisecond));
    }

    @Test
    public void shouldBeInstalledASensorToAUserRamclen() {
        Watcher watcher = new Watcher("objectId", "measurableAttributeName");
        state().watcherContainer().addWatcher("Ramclen", watcher);
        HashMap<String, List<Watcher>> watchers = state().watcherContainer().getWatchers();
        assertNotNull(watchers.get("Ramclen"));
        assertSame(watcher, watchers.get("Ramclen").get(0));
    }

    @Test
    public void shouldReturnPausedSimulationState() {
        state().setSimulationState(SimulationState.PAUSE);
        assertSame(state().getSimulationState(), SimulationState.PAUSE);
    }

    @Test
    public void shouldFiredTheListenerWhenChangeTheSimulationState() {
        SimulationStateListener listener = mock(SimulationStateListener.class);
        state().setSimulationStateListener(listener);
        state().setSimulationState(SimulationState.PAUSE);
        verify(listener).change(SimulationState.PAUSE);
    }

    private class MySimulationAgent implements SimulationAgent {

        @Override
        public List<String> roots() {
            LinkedList<String> list = new LinkedList<>();
            list.add("happy-building");
            list.add("bored-building");
            list.add("unhappy-building");
            return list;
        }

        @Override
        public List<String> childrenOf(String objectId) {
            LinkedList<String> list = new LinkedList<>();
            list.add("room1");
            list.add("room2");
            list.add("room3");
            return list;
        }

        @Override
        public List<String> attributesOf(String objectId) {
            LinkedList<String> list = new LinkedList<>();
            list.add("sensor1");
            list.add("sensor2");
            list.add("sensor3");
            return list;
        }

        @Override
        public String valueGet(String objectId) {
            return null;
        }

        @Override
        public boolean valueSet(String objectId, String value) {
            return false;
        }

        @Override
        public boolean play() {
            return false;
        }

        @Override
        public boolean pause() {
            return false;
        }

        @Override
        public boolean reset() {
            return false;
        }
    }

}
